#!/usr/bin/env python
# *-* coding:UTF-8 *-*
#
#  pymusic.py
#
#       Stephen Smally <stephen.smally@gmail.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#       
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#       
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#       


'''Simple Music Player written in pygtk using gstreamer'''

import gtk
import gst
import os
import shelve
import pango
import widgets

global userhome
userhome = os.getenv("HOME")

global theme
theme = '''
style "pymusic" {
    GtkPaned::handle-size = 1
    GtkPaned::handle-type = 0
    engine "murrine" {
    separatorstyle = 0
    }
}

style "bookmarks" {
    GtkTreeView::odd-row-color = @bg_color
    GtkTreeView::even-row-color = @bg_color
}

widget_class "GtkWindow*" style "pymusic"
widget "*bookmarks*" style "bookmarks"
'''

class PyMusicView(gtk.Window):
    def __init__(self):
        '''Setting up the view'''
        super(PyMusicView, self).__init__()
        self.set_title("PyMusic")
        gtk.rc_parse_string(theme)
        try:
            self.set_icon_from_file("/usr/share/pixmaps/pymusic.png")
        except: pass
        self.set_position(gtk.WIN_POS_CENTER)
        self.set_default_size(800, 500)
        self.realize()
        self.window_style = self.get_style()
        self.add_widgets()
        self.show_all()
        self.separator.set_visible(False)
        for items in self.main_toolbar.get_children():
            if not items in [self.refresh, self.add_fold]:
                items.set_sensitive(False)
        self.bookmarks_list.set_visible(False)
        self.library_delete.set_visible(False)


    def create_toolbar(self):
        self.refresh = gtk.ToolButton(gtk.STOCK_REFRESH)
        self.refresh.set_tooltip_text("Refresh Songs/Artists lists")
        self.add_fold = gtk.ToolButton(gtk.STOCK_ADD)
        self.add_fold.set_tooltip_text("Add a folder containg music")
        self.previous = gtk.ToolButton(gtk.STOCK_MEDIA_PREVIOUS)
        self.previous.set_tooltip_text("Previous song")
        self.play_pause = gtk.ToolButton(gtk.STOCK_MEDIA_PLAY)
        self.play_pause.set_tooltip_text("Play or pause the song")
        self.stop = gtk.ToolButton(gtk.STOCK_MEDIA_STOP)
        self.stop.set_tooltip_text("Stop the song")
        self.next = gtk.ToolButton(gtk.STOCK_MEDIA_NEXT)
        self.next.set_tooltip_text("Next song")
        self.search_tool = widgets.SearchEntry()
        self.active_song = gtk.Label("Title - Artist - Album")
        self.active_song.set_ellipsize(pango.ELLIPSIZE_END)
        self.active_song.set_tooltip_text("Current song (Title - Artist - Album)")
        self.active_song_cont = gtk.ToolItem()
        self.active_song_cont.add(self.active_song)
        self.active_song_cont.set_expand(True)
        self.main_toolbar = gtk.Toolbar()
        self.main_toolbar.insert(self.refresh, -1)
        self.main_toolbar.insert(self.add_fold, -1)
        self.main_toolbar.insert(self.previous, -1)
        self.main_toolbar.insert(self.play_pause, -1)
        self.main_toolbar.insert(self.stop, -1)
        self.main_toolbar.insert(self.next, -1)
        self.main_toolbar.insert(self.active_song_cont, -1)
        self.main_toolbar.insert(self.search_tool, -1)
        return self.main_toolbar
    
    def create_song_list(self):
        self.song_store = gtk.ListStore(str, str, str, str, str, 'PangoEllipsizeMode')
        self.song_list = gtk.TreeView(self.song_store)
        self.playing_renderer = gtk.CellRendererPixbuf()
        self.playing_column = gtk.TreeViewColumn("Title", self.playing_renderer)
        self.playing_column.add_attribute(self.playing_renderer, "stock-id", 0)
        self.title_renderer = gtk.CellRendererText()
        self.title_column = gtk.TreeViewColumn("Title", self.title_renderer)
        self.title_column.add_attribute(self.title_renderer, "text", 2)
        self.title_column.add_attribute(self.title_renderer, "ellipsize", 5)
        self.title_column.set_resizable(True)
        self.title_column.set_expand(True)
        self.song_list.append_column(self.title_column)
        self.artist_renderer = gtk.CellRendererText()
        self.artist_column = gtk.TreeViewColumn("Artist", self.artist_renderer)
        self.artist_column.add_attribute(self.artist_renderer, "text", 3)
        self.artist_column.add_attribute(self.artist_renderer, "ellipsize", 5)
        self.artist_column.set_resizable(True)
        self.artist_column.set_expand(True)
        self.song_list.append_column(self.artist_column)
        self.album_renderer = gtk.CellRendererText()
        self.album_column = gtk.TreeViewColumn("Album", self.album_renderer)
        self.album_column.add_attribute(self.album_renderer, "text", 4)
        self.album_column.add_attribute(self.album_renderer, "ellipsize", 5)
        self.album_column.set_resizable(True)
        self.album_column.set_expand(True)
        self.song_list.append_column(self.album_column)
        self.song_list.set_rules_hint(True)
        self.song_list.set_expander_column(self.title_column)
        self.song_cont = gtk.ScrolledWindow()
        self.song_cont.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.song_cont.add(self.song_list)
        return self.song_cont
    
    def create_artist_list(self):
        self.artist_store = gtk.TreeStore(str)
        self.artist_list = gtk.TreeView(self.artist_store)
        self.artist_list_renderer = gtk.CellRendererText()
        self.artist_list_column = gtk.TreeViewColumn("Artist", self.artist_list_renderer)
        self.artist_list_column.add_attribute(self.artist_list_renderer, "text", 0)
        self.artist_list.append_column(self.artist_list_column)
        self.artist_list.set_rules_hint(True)
        self.artist_cont = gtk.ScrolledWindow()
        self.artist_cont.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.artist_cont.add(self.artist_list)
        return self.artist_cont
    
    def create_libraries_list(self):
        self.library_store = gtk.ListStore(str, str, str)
        self.library_list = gtk.TreeView(self.library_store)
        self.library_renderer = gtk.CellRendererText()
        self.icon_renderer = gtk.CellRendererPixbuf()
        self.library_column = gtk.TreeViewColumn("Folder")
        self.library_column.pack_start(self.icon_renderer, False)
        self.library_column.pack_start(self.library_renderer)
        self.library_column.add_attribute(self.icon_renderer, "icon-name", 0)
        self.library_column.add_attribute(self.library_renderer, "text", 1)
        self.library_column.set_expand(True)
        self.library_2_renderer = gtk.CellRendererText()
        self.library_2_column = gtk.TreeViewColumn("Library", self.library_2_renderer)
        self.library_2_column.add_attribute(self.library_2_renderer, "text", 2)
        self.library_2_column.set_expand(True)
        self.library_list.append_column(self.library_column)
        self.library_list.append_column(self.library_2_column)
        self.library_list.set_rules_hint(True)
        self.library_cont = gtk.ScrolledWindow()
        self.library_cont.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.library_cont.add(self.library_list)
        self.library_delete = gtk.Button("")
        self.library_delete.set_border_width(1)
        self.library_box = gtk.VBox()
        self.library_box.pack_start(self.library_cont)
        self.library_box.pack_start(gtk.HSeparator(), False)
        self.library_box.pack_start(self.library_delete, False)
        return self.library_box
    
    def create_start_diag(self):
        self.start_diag = gtk.Fixed()
        self.start_label = gtk.Label("Insert a folder contains music or lib path")
        self.help_button = gtk.Button("", gtk.STOCK_HELP)
        self.start_combo = gtk.combo_box_entry_new_text()
        self.start_combo.set_usize(290, 28)
        self.start_entry = self.start_combo.child
        self.start_button = gtk.Button("Import Music")
        self.start_store = self.start_combo.get_model()
        self.start_diag.put(self.start_label, 285, 155)
        self.start_diag.put(self.start_combo, 262, 175)
        self.start_diag.put(self.start_button, 355, 205)
        self.start_diag.put(self.help_button, 370, 305)
        return self.start_diag
    
    def create_misc(self):
        self.songs_count = gtk.Label("Songs")
        self.separator = gtk.VSeparator()
    
    def create_bookmarks(self):
        self.bookmarks_store = gtk.TreeStore(str, str, int)
        self.bookmarks_list = gtk.TreeView(self.bookmarks_store)
        self.bookmarks_renderer = gtk.CellRendererPixbuf()
        self.bookmarks_text = gtk.CellRendererText()
        self.bookmarks_column = gtk.TreeViewColumn("Music")
        self.bookmarks_column.pack_start(self.bookmarks_renderer, False)
        self.bookmarks_column.pack_start(self.bookmarks_text, False)
        self.bookmarks_column.add_attribute(self.bookmarks_renderer, "icon-name", 0)
        self.bookmarks_column.add_attribute(self.bookmarks_renderer, "stock-size", 2)
        self.bookmarks_column.add_attribute(self.bookmarks_text, "text", 1)
        self.parent_music = self.bookmarks_store.append(None, ["emblem-sound", "Music", 2])
        self.parent_datas = self.bookmarks_store.append(None, ["emblem-system", "Datas", 2])
        for items in [["audio-mpeg", "Songs"], ["stock_person", "Artists"]]:
            self.bookmarks_store.append(self.parent_music, [items[0], items[1], 2])
        self.bookmarks_store.append(self.parent_datas, ["text-x-generic", "Libs", 2])
        self.bookmarks_list.set_headers_visible(False)
        self.bookmarks_list.append_column(self.bookmarks_column)
        self.bookmarks_list.set_level_indentation(-5)
        self.bookmarks_list.set_name("bookmarks")
        return self.bookmarks_list
    
    def add_widgets(self):
        self.create_misc()
        self.main_vbox = gtk.VBox()
        self.main_hbox = gtk.HBox()
        self.main_vbox.pack_start(self.create_toolbar(), False)
        self.var_cont = gtk.Notebook()
        self.var_cont.set_show_tabs(False)
        self.var_cont.set_show_border(False)
        self.var_workaround = gtk.HBox()
        self.var_workaround.pack_start(self.separator, False)
        self.var_workaround.pack_start(self.var_cont)
        self.main_vbox.pack_start(self.main_hbox)
        self.main_hbox.pack_start(self.create_bookmarks(), False, False)
        self.main_hbox.pack_start(self.var_workaround)
        self.var_cont.append_page(self.create_start_diag())
        self.var_cont.append_page(self.create_song_list())
        self.var_cont.append_page(self.create_artist_list())
        self.var_cont.append_page(self.create_libraries_list())
        self.main_vbox.pack_start(gtk.HSeparator(), False)
        self.main_vbox.pack_start(self.songs_count, False)
        self.add(self.main_vbox)

#-------------------------------------------------------------
#
#-------------------------------------------------------------

class PyMusicDatabase:
    def __init__(self, lib):
        self.lib = lib
        print(" ".join(["opening", self.lib]))
        self.library = shelve.open(self.lib, writeback=True)
    
    def get_key(self, key):
        self.key = self.library[key]
        return self.key
    
    def get_all(self):
        self.all = []
        for items in sorted(self.library):
            self.all.append(items)
        return self.all
    
    def write_key(self, key, value):
        self.library[key] = value
    
    def close(self):
        self.library.close()
        print(" ".join(["closing", self.lib]))
        

#-------------------------------------------------------------
#
#-------------------------------------------------------------

class PyMusicControl:
    def __init__(self):
        '''Setting up the Player'''
        self.view = PyMusicView()
        self.settings = shelve.open("/".join([userhome, ".pymusic.conf"]), writeback=True)
        for items in self.settings.keys():
            self.view.library_store.append(["folder", self.settings[items], items])
            self.view.start_store.append([items])
        if len(self.view.start_store) == 0:
            self.view.start_store.append(["You havent any lib yet"])
        self.last_played = 0
        #-----------------------------------------------------
        self.player = gst.element_factory_make("playbin2", "PyMusic player")
        self.sink = gst.element_factory_make("autoaudiosink", "PyMusic Output")
        self.videosink= gst.element_factory_make("fakesink", "PyMusic fake Output")
        self.player.set_property("video-sink", self.videosink)
        self.player.set_property("audio-sink", self.sink)
        self.player_bus = self.player.get_bus()
        self.player_bus.add_signal_watch()
        #-----------------------------------------------------
        self.player_bus.connect("message", self.On_messages)
        self.view.start_button.connect("clicked", self.startup_button)
        self.view.help_button.connect("clicked", self.On_help)
        self.view.refresh.connect("clicked", lambda w: self.work_library())
        self.view.add_fold.connect("clicked", self.On_add_music)
        self.view.song_list.connect("row-activated", self.On_activated)
        self.view.stop.connect("clicked", self.On_stop)
        self.view.play_pause.connect("clicked", self.On_media_button)
        self.view.next.connect("clicked", self.On_next)
        self.view.previous.connect("clicked", self.On_previous)
        self.view.bookmarks_list.connect("cursor-changed", self.change_tab)
        self.view.library_list.connect("cursor-changed", self.select_lib)
        self.view.library_delete.connect("clicked", self.delete_lib)
        self.view.search_tool.search_function = self.search
        self.view.connect("size-allocate", self.On_refresh)
        self.view.connect("destroy", self.On_exit)
        #-----------------------------------------------------
    
    def search(self, string):
        if not string == "":
            self.song_store_search = gtk.ListStore(str, str, str, str, str, 'PangoEllipsizeMode')
            for items in self.view.song_store:
                if string.lower() in items[2].lower():
                    self.song_store_search.append([items[0], items[1], items[2], items[3], items[4], items[5]])
            self.view.song_list.set_model(self.song_store_search)
        else:
            self.view.song_list.set_model(self.view.song_store)
    
    def change_tab(self, widget):
        self.tabs_dict = {
        "Songs": 1,
        "Artists": 2,
        "Libs": 3
        }
        try:
            self.select = self.view.bookmarks_store[self.view.bookmarks_list.get_selection().get_selected()[1]][1]
        except TypeError:
            self.select = None
        if not self.select in ["Music", "Datas", None]:
            if self.select in ["Artists", "Libs"]:
                self.view.search_tool.set_sensitive(False)
            else:
                self.view.search_tool.set_sensitive(True)
            self.view.var_cont.set_page(self.tabs_dict[self.select])
    
    def On_help(self, widget):
        self.info_message = '''Insert in the entry the path to the folder that contains your
music (e.g. /home/user/Music), or a path to an existent
library generated by PyMusic.
You can see your libs clicking on arrow of the combobox.'''
        self.info_dialog = gtk.MessageDialog(self.view, 0, gtk.MESSAGE_INFO, gtk.BUTTONS_OK, "What i have to do now?")
        self.info_dialog.format_secondary_text(self.info_message)
        self.info_dialog.set_title("Infos on PyMusic")
        self.info_dialog.run()
        self.info_dialog.destroy()
    
    def startup_button(self, widget):
        try:
            self.startup_collection(widget, "/".join([self.view.start_entry.get_text().split("/.library.dat")[0], ".library.dat"]))
        except:
            self.dialog = gtk.MessageDialog(self.view, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, "Error!")
            self.dialog.format_secondary_text("Permission Denied")
            self.dialog.run()
            self.dialog.destroy()
    
    def startup_collection(self, widget, library):
        self.folder_library = library
        self.folder = self.folder_library.replace("/.library.dat", "")
        self.view.song_store.clear()
        self.view.artist_store.clear()
        if os.path.exists(self.folder):
            self.settings[self.folder_library] = self.folder
            if not os.path.exists(self.folder_library):
                print("Creating a new library...")
                self.create_lib()
            print("Using the pre-existing library...")
            self.work_library()
        else:
            self.view.songs_count.set_text("Can't find the folder!")
        self.view.start_entry.set_text("")
    
    def work_library(self):
        self.database = PyMusicDatabase(self.folder_library)
        self.view.song_store.clear()
        self.view.artist_store.clear()
        for items in self.database.get_all():
            self.view.song_store.append([
            None,
            items,
            self.database.get_key(items)[0],
            self.database.get_key(items)[1],
            self.database.get_key(items)[2],
            pango.ELLIPSIZE_END
            ])
        self.artist_dict = []
        self.album_dict = []
        for items in self.database.get_all():
            if not self.database.get_key(items)[1] in self.artist_dict:
                self.artist_dict.append(self.database.get_key(items)[1])
        for items in self.artist_dict:
            self.parent_artist = self.view.artist_store.append(None, [items])
            for item in self.database.get_all():
                if self.database.get_key(item)[1] == items:
                    if not self.database.get_key(item)[2] in self.album_dict:
                        self.parent_album = self.view.artist_store.append(self.parent_artist, [self.database.get_key(item)[2]])
                        self.album_dict.append(self.database.get_key(item)[2])
                        for songs in self.database.get_all():
                            if self.database.get_key(songs)[2] == self.database.get_key(item)[2]:
                                self.view.artist_store.append(self.parent_album, [self.database.get_key(songs)[0]])
        self.view.start_store.clear()
        self.view.library_store.clear()
        for items in self.settings.keys():
            self.view.library_store.append(["folder", self.settings[items], items])
            self.view.start_store.append([items])
        self.view.songs_count.set_text(" ".join([str(len(self.database.get_all())), "songs in this library"]))
        self.view.bookmarks_list.set_visible(True)
        self.view.var_cont.set_page(1)
        self.view.separator.set_visible(True)
        for items in self.view.main_toolbar.get_children():
            items.set_sensitive(True)
        self.select(0)
        self.database.close()
        
    def create_lib(self):
        '''set up the collection'''
        self.database = PyMusicDatabase(self.folder_library)
        import mutagen.mp3
        self.files = []
        for files in os.walk(self.folder):
            for item in files[2]:
                self.file = "/".join([files[0], item])
                if os.path.exists(self.file):
                    if item.split(".")[-1] in ["mp3", "ogg", "m4a"]:
                        self.files.append(self.file)
        #-----------------------------------------------------
        for items in self.files:
            try:
                self.media = mutagen.mp3.Open(items)
                self.database.write_key(items, [
                str(self.media["TIT2"]),
                str(self.media["TPE1"]),
                str(self.media["TALB"])
                ])
            except:
                self.database.write_key(items, [
                items.split("/")[-1],
                "Unknown Artist",
                "Unknown Album"
                ])
        #-----------------------------------------------------
        self.database.close()

    def On_messages(self, bus, message):
        if message.type == gst.MESSAGE_EOS:
            self.On_next(None)
        elif message.type == gst.MESSAGE_ERROR:
            print(message)

    def On_media_button(self, widget):
        if widget.get_stock_id() == gtk.STOCK_MEDIA_PLAY:
            widget.set_stock_id(gtk.STOCK_MEDIA_PAUSE)
            if self.player.get_property("uri") != None:
                self.player.set_state(gst.STATE_PLAYING)
            else:
                self.On_play(0, 0)
        #-----------------------------------------------------
        elif widget.get_stock_id() == gtk.STOCK_MEDIA_PAUSE:
            widget.set_stock_id(gtk.STOCK_MEDIA_PLAY)
            self.player.set_state(gst.STATE_PAUSED)

    def On_play(self, path, genre):
        '''Play or stop the song'''
        self.player.set_state(gst.STATE_READY)
        if genre == 0:
            self.data = self.view.song_store[path]
        if genre == 1:
            self.data = self.view.artist_store[path]
        self.view.active_song.set_text(" - ".join([self.data[2], self.data[3], self.data[4]]))
        self.player.set_property("uri", "file://"+self.data[1])
        self.player.set_state(gst.STATE_PLAYING)
        self.view.play_pause.set_stock_id(gtk.STOCK_MEDIA_PAUSE)
        self.last_played = path
    
    def On_stop(self, widget):
        self.player.set_state(gst.STATE_READY)
        self.view.play_pause.set_stock_id(gtk.STOCK_MEDIA_PLAY)
        
    def On_activated(self, widget, path, view):
        self.On_play(path[0], 0)

    def On_next(self, widget):
        if self.last_played != len(self.view.song_store)-1:
            self.On_play(self.last_played+1, 0)
        else:
            self.On_play(self.last_played, 0)
        
    def On_previous(self, widget):
        if self.last_played != 0:
            self.On_play(self.last_played-1, 0)
        else:
            self.On_play(self.last_played, 0)

    def On_refresh(self, widget, rec):
        self.size = widget.get_size()[0]/3
        self.view.title_column.set_max_width(self.size)
        self.view.artist_column.set_max_width(self.size)
        self.view.album_column.set_max_width(self.size)

    def select(self, selection):
        self.view.var_cont.set_page(selection+1)

    def On_add_music(self, widget):
        self.view.start_store.clear()
        self.view.library_store.clear()
        for items in self.settings.keys():
            self.view.library_store.append(["folder", self.settings[items], items])
            self.view.start_store.append([items])
        if len(self.view.start_store) == 0:
            self.view.start_store.append(["You havent any lib yet"])
        self.view.var_cont.set_page(0)
        self.view.separator.set_visible(False)
        self.view.bookmarks_list.set_visible(False)
    
    def delete_lib(self, widget):
        try:
            self.lib_to_delete = self.view.library_store[self.view.library_list.get_selection().get_selected()[1]][2]
        except IndexError:
            self.lib_to_delete = "The lib doesn't exists"
        if os.path.exists(self.lib_to_delete):
            try:
                os.remove(self.lib_to_delete)
                del self.settings[self.lib_to_delete]
                self.settings.close()
                self.settings = shelve.open("/".join([userhome, ".pymusic.conf"]), writeback=True)
                self.On_add_music(None)
                widget.hide()
            except OSError:
                self.dialog = gtk.MessageDialog(self.view, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, "Error!")
                self.dialog.format_secondary_text("Permission Denied")
                self.dialog.run()
                self.dialog.destroy()
        else:
            print("The lib doesn't exists")
    
    def select_lib(self, widget):
        self.lib_to_delete = self.view.library_store[self.view.library_list.get_selection().get_selected()[1]][2]
        self.view.library_delete.set_label("Delete the '%s' library" % self.lib_to_delete)
        self.view.library_delete.show()
    
    def On_exit(self, root):
        print("The window %s is going to quit" % root)
        self.player.set_state(gst.STATE_NULL)
        try:
            self.database.close()
        except: pass
        self.settings.close()
        gtk.main_quit()


if __name__ == "__main__":
    app = PyMusicControl()
    gtk.main()
